/*		 ITI-FG  Kernel module free public release
 *
 *		     Copyright (C) 1996-2000 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: itifgAux.c,v 7.0 2000/10/25 09:07:57 cvs Exp $
 *
 */

/*
 * Includes:
 */

#include "itifgOS.h"	/* os specific stuff */
#include "itifgExt.h"	/* offical driver interface */
#include "itifgReg.h"	/* register definitions ICR Regs */
#include "icpciReg.h"	/* register definitions IC-PCI */
#include "amvsReg.h"	/* register definitions AM-VS */
#include "amdigReg.h"	/* register definitions AM-DIG */
#include "ampcvReg.h"	/* register definitions PCVision */
#include "pcdigReg.h"	/* register definitions PC-DIG */
#include "itifgIntD.h"	/* driver internals */
#include "itifgIntF.h"	/* internal functions */

/*
 * Defines:
 */

/*
 * Globals:
 */

/* human readable error messages */
const char *iti_errmsg[] =
{ "no error",
  "mask error",	
  "range error",
  "operation not supported",
  "null pointer access",
  "operation already in use"
};

/* the possible modules and their names */
/* this have to be the same as in procfs.c */
const struct mod_names_t mod_names[] =
{
  { ICP_AMVS, "AM-VS" },
  { ICP_AMDG, "AM-DIG" },
  { ICP_AMPV, "PCVision" },
  { ITI_VMPCD, "PCDig" }
};

extern struct vs_cam_t vs_def_cam;	/* \               */
extern struct dg_cam_t dg_def_cam;	/* - see itifgAM.c */
extern struct pv_cam_t pv_def_cam;	/* /               */
extern struct pcd_cam_t pcd_def_cam;    /* /		   */

/*
 * Prototypes:
 */

/*
 * iti_error
 */

void
iti_error (itifg_t *iti, const char* function)
{
  iti_printna ("%s: %s\n", function, iti_errmsg[-iti->errno]);
  iti->errno = 0;
}

/*
 *  iti_decode_devno
 */

int
iti_decode_devno (iti_dev_t dev, int *ret_board, int *ret_funct)
{
  int board, funct, major, minor;

  iti_get_major (dev, &major);
  iti_get_minor (dev, &minor);

  /* decode */
  board = (minor & ITI_BOARD_MASK) >> ITI_BOARD_SHIFT;
  funct = (minor & ITI_FUNCT_MASK) >> ITI_FUNCT_SHIFT;
  
  /* device OK? */
  /* board request valid? */
  if (board >= boards)
    return -ENXIO;

  /* functionality request valid? */
  switch (funct)
    {
    case ITI_FUNCT_DATA:
    case ITI_FUNCT_PORT:
      break;
    default:
      return -ENXIO;
    }
  
  if (ret_board)
    *ret_board = board;
  if (ret_funct)
    *ret_funct = funct;

  return OK;
}

/*
 * New Interrupt Handler Interface
 */

int
iti_enable_irq (itifg_t *iti, iti_eirq_t mode)
{
  u_long irqs = 0UL;

  ITI_PRINT1("> iti_enable_irq\n");

  /* disable irq's */  
  iti_irq_lock (iti);
  /* clear all requests */
  xcr_test_irqs (iti, &irqs);
  if (irqs)
    {
      ITI_PRINT1A("enable_irq begin (%lx)\n", irqs);
      xcr_clear_irqs (iti);
    }

  switch (mode)
    {
    case ITI_EIRQ_MAIN_OPEN:
      CHECK_FOR_EIO (icr_set_inten, ICR_ENABLE);
      if (IS_ICPCI)
	{
	  CHECK_FOR_EIO (icr_set_binten, ICR_ENABLE);
	}
      if (IS_PCDIG)
	{
	  CHECK_FOR_EIO (pcd_set_bmien, PCD_ENABLE);
	}
      break;
    case ITI_EIRQ_DATA_OPEN:
      switch (DEVICE_ID)
	{
	case ITI_ICPCI:
	  if (iti->get_smstat (iti) == CR_ILACE)
	    {
	      CHECK_FOR_EIO (cr_set_vbien, CR_DISABLE);
	      CHECK_FOR_EIO (cr_set_evbien, CR_ENABLE);
	      CHECK_FOR_EIO (cr_set_ovbien, CR_DISABLE);
	    }
	  else
	    {
	      CHECK_FOR_EIO (cr_set_vbien, CR_ENABLE);
	      CHECK_FOR_EIO (cr_set_evbien, CR_DISABLE);
	      CHECK_FOR_EIO (cr_set_ovbien, CR_DISABLE);
	    }
	  break;
	case ITI_PCDIG:
	  CHECK_FOR_EIO (pcd_set_vbien, CR_ENABLE);
	  break;
	default:
	  return -ENXIO;
	}
      CHECK_FOR_EIO (iti->set_adrien, CR_ENABLE);
      break;
    case ITI_EIRQ_PORT_OPEN:
      if (IS_REV_B || IS_PCDIG)
	{
	  CHECK_FOR_EIO (iti->set_inpien, CR_ENABLE);
	}
      break;
    default:
      return -EINVAL;
    }

  xcr_test_irqs (iti, &irqs);
  if (irqs)
    {
      ITI_PRINT1A("enable_irq end (%lx)\n", irqs);
      xcr_clear_irqs (iti);
    }
  /* enable irq's */
  iti_irq_unlock (iti);

  ITI_PRINT1("< iti_enable_irq\n");
  return OK;
}

int
iti_disable_irq (itifg_t *iti, iti_dirq_t mode)
{
  u_long irqs = 0UL;

  ITI_PRINT1("> iti_disable_irq\n");

  /* disable irq's */
  iti_irq_lock (iti);
  /* clear all requests */
  xcr_test_irqs (iti, &irqs);
  if (irqs)
    {
      ITI_PRINT1A("disable_irq begin (%lx)\n", irqs);
      xcr_clear_irqs (iti);
    }

  switch (mode)
    {
    case ITI_DIRQ_MAIN_CLOSE:
      if (IS_ICPCI)
	{
	  CHECK_FOR_EIO (icr_set_binten, ICR_DISABLE);
	}
      if (IS_PCDIG)
	{
	  CHECK_FOR_EIO (pcd_set_bmien, PCD_DISABLE);
	}
      CHECK_FOR_EIO (icr_set_inten, ICR_DISABLE);
      break;
    case ITI_DIRQ_DATA_CLOSE:
      CHECK_FOR_EIO (iti->set_adrien, CR_ENABLE);
      CHECK_FOR_EIO (iti->set_vbien, CR_DISABLE);
      if (IS_ICPCI)
	{
	  CHECK_FOR_EIO (cr_set_evbien, CR_DISABLE);
	  CHECK_FOR_EIO (cr_set_ovbien, CR_DISABLE);
	}
      break;
    case ITI_DIRQ_PORT_CLOSE:
      if (IS_REV_B || IS_PCDIG)
	{
	  CHECK_FOR_EIO (iti->set_inpien, CR_DISABLE);
	}
      break;
    default:
      return -EINVAL;
    }

  /* clear all requests */
  xcr_test_irqs (iti, &irqs);
  if (irqs)
    {
      ITI_PRINT1A("disable_irq end (%lx)\n", irqs);
      xcr_clear_irqs (iti);
    }
  /* enable irq's */
  iti_irq_unlock (iti);

  ITI_PRINT1("< iti_disable_irq\n");
  return (OK);
}

/*
 * iti_icr_init()
 */

static int
iti_icr_init (itifg_t *iti)
{
  ITI_PRINT1("> iti_icr_init\n");

  iti->errno = EITI_ERR_NO;

  /* initialize registers */
  /* icr_set_mbox1 (iti, ICR_MBOX_INIT); */ /* board used marker */
  icr_set_mbox2 (iti, ICR_MBOX_INIT);
  icr_set_mbox3 (iti, ICR_MBOX_INIT);
  icr_set_mbox4 (iti, ICR_MBOX_INIT);
  if (IS_ICPCI)
    {
      icr_set_bmdst (iti, ICR_BMDST_INIT);
      icr_set_bmxc (iti, ICR_BMXC_INIT);
    }
  icr_init_intctl (iti);
  icr_init_bmctl (iti);

  /* check for errors */
  if (iti->errno < 0)
    {
      iti_error (iti, "iti_icr_init");
      return -EIO;
    }

  ITI_PRINT1("< iti_icr_init\n");
  return OK;
}

/*
 * iti_cr_init()
 */

int
iti_cr_init (itifg_t *iti)
{
  ITI_PRINT1("> iti_cr_init\n");

  iti->errno = EITI_ERR_NO;

  /* initialize registers */
  switch (DEVICE_ID)
    {
    case ITI_ICPCI:
      cr_init_con (iti);
      cr_init_acqcon (iti);
      cr_set_pitch (iti, CR_PITCH_INIT);
      cr_set_aoix (iti, CR_AOIX_INIT);
      cr_set_bmsf (iti, CR_BMSF_INIT);
      cr_set_bmss (iti, CR_BMSS_INIT);
      cr_set_aqstart (iti, CR_AQSTART_INIT);
      cr_set_intadr (iti, CR_INTADR_INIT);
      cr_init_intcon (iti);
      cr_init_mdata (iti);
      /* extended init on REV-B boards (including PCVision) */
      if (IS_REV_B)
	{
	  cr_init_wzoom (iti);
	  cr_init_portcon (iti);
	  cr_set_outport (iti, CR_OUTPORT_INIT);
	}
      /* set right memory size */
      if (iti->big_mem)
	  cr_set_memcfg (iti, CR_MEM_4);
      else
	  cr_set_memcfg (iti, CR_MEM_2);
      /* set the virtual switch table functions */
      iti->get_intadr = cr_get_intadr;
      iti->set_intadr = cr_set_intadr;
      iti->get_acqmd = cr_get_acqmd;
      iti->set_acqmd = cr_set_acqmd;
      iti->get_aqadr = cr_get_aqadr;
      iti->get_pixsz = cr_get_pixsz;
      iti->get_hdec = cr_get_hdec;
      iti->set_hdec = cr_set_hdec;
      iti->get_vdec = cr_get_vdec;
      iti->set_vdec = cr_set_vdec;
      iti->set_fldsel = cr_set_fldsel;
      iti->get_smstat = cr_get_smstat;
      iti->set_vbien = cr_set_vbien;  
      iti->set_adrien = cr_set_adrien;  
      iti->set_outstrb = cr_set_outstrb;
      iti->set_inregenb = cr_set_inregenb;
      iti->set_instrbpol = cr_set_instrbpol;
      iti->set_inregclr = cr_set_inregclr;
      iti->set_inpintpol = cr_set_inpintpol;
      iti->set_inpien = cr_set_inpien;
      iti->get_inport = cr_get_inport;
      iti->set_outport = cr_set_outport;
      break;

    case ITI_PCDIG:

      /* stop here, if not initialized */
      if (!FPGA_OK)
	return OK;
      
      /* busmaster and interrupt registes */
      pcd_init_bmcon (iti);
      pcd_init_bmlut (iti);
      pcd_set_aqstart (iti, PCD_AQSTART_INIT);
      pcd_set_intadr (iti, PCD_INTADR_INIT);
      pcd_set_segsize (iti, PCD_SEGSIZE_INIT);
      pcd_set_octstart (iti, PCD_OCTSTART_INIT);
      pcd_set_swcount (iti, PCD_SWCOUNT_INIT);
      pcd_init_zoom (iti);
      pcd_init_intcon1 (iti);
      /* camera and acquisition registers */
      pcd_init_acqcon (iti);
      pcd_init_intcon0 (iti);
      pcd_init_decim (iti);
      pcd_init_portcon (iti);
      pcd_set_outport (iti, PCD_OUTPORT_INIT);
      /* OCT/DMA/LUT0/LUT1 Table */
      pcd_set_octacc (iti, PCD_ENABLE);
      iti_memset_io (iti->pcd_oct_table, PCD_RES_INIT,
		     PCD_NUM_OCT_ENTRYS);
      pcd_set_octacc (iti, PCD_DISABLE);
      pcd_set_dmaacc (iti, PCD_ENABLE);
      iti_memset_io (iti->pcd_dma_table, PCD_RES_INIT,
		     PCD_NUM_DMA_ENTRYS - 2);
      pcd_set_dmaacc (iti, PCD_DISABLE);
      pcd_set_lutacc (iti, PCD_ENABLE);
      iti_memset_io (iti->pcd_lut0_table, PCD_RES_INIT,
		     PCD_NUM_LUT_ENTRYS);
      iti_memset_io (iti->pcd_lut1_table, PCD_RES_INIT,
		     PCD_NUM_LUT_ENTRYS);
      pcd_set_lutacc (iti, PCD_DISABLE);
      /* set the virtual switch table functions */
      iti->get_intadr = pcd_get_intadr;
      iti->set_intadr = pcd_set_intadr;
      iti->get_acqmd = pcd_get_acqmd;
      iti->set_acqmd = pcd_set_acqmd;
      iti->get_aqadr = pcd_get_aqadr;
      iti->get_pixsz = pcd_get_pixsz;
      iti->get_hdec = pcd_get_hdec;
      iti->set_hdec = pcd_set_hdec;
      iti->get_vdec = pcd_get_vdec;
      iti->set_vdec = pcd_set_vdec;
      iti->set_fldsel = pcd_set_fldsel;
      iti->get_smstat = pcd_get_smstat;
      iti->set_vbien = pcd_set_vbien;  
      iti->set_adrien = pcd_set_adrien;  
      iti->set_outstrb = pcd_set_outstrb;
      iti->set_inregenb = pcd_set_inregenb;
      iti->set_instrbpol = pcd_set_instrbpol;
      iti->set_inregclr = pcd_set_inregclr;
      iti->set_inpintpol = pcd_set_inpintpol;
      iti->set_inpien = pcd_set_pioien;
      iti->get_inport = pcd_get_inport;
      iti->set_outport = pcd_set_outport;
      break;

    default:
      return -ENXIO;
    }

  /* check for errors */
  if (iti->errno < 0)
    {
      iti_error (iti, "iti_cr_init");
      return -EIO;
    }

  ITI_PRINT1("< iti_cr_init\n");
  return OK;
}

/*
 * iti_pci_init
 */

int
iti_pci_init (itifg_t *iti, void *data)
{
  int error;
  
  ITI_PRINT1("> iti_pci_init\n");
  
  if ((error = iti_os_pci_init (iti, data)))
    return error;
  
  /* some informational and debugging output of the pci config space */
  switch (iti->device_id)
    {
    case ITI_PCI_DEVICE_ID_IMAGINGTECH_ICPCI:
      iti_printi ("IC-PCI frame grabber board found.\n");
      break;
    case ITI_PCI_DEVICE_ID_IMAGINGTECH_ASYNC:
      iti_printn ("IC-ASYNC frame grabber board found.\n");
      iti_printn ("Not supported yet!\n");
      return -ENXIO;
      break;
    case ITI_PCI_DEVICE_ID_IMAGINGTECH_PCDIG:
      iti_printi ("PC-DIG frame grabber board found.\n");
      break;
    default:
      iti_printna ("Unknown frame grabber board found (%d).\n",
		   iti->device_id);
      return -ENXIO;	  
    }
  
  if (iti->irq_line == 0 || iti->irq_line == 255)
    {
      iti->irq_line = 255;
      iti_printw ("interrupt disabled!\n");
      /* return -EINVAL */
    }
  else
    {
      if (iti->irq_line < 3 || iti->irq_line > 31 )
	{
	  iti_printea ("really use interrupt %d?\n", iti->irq_line);
	  return -EINVAL;
	}
    }
  
  ITI_PRINT1A("    Interrupt : %d\n", iti->irq_line);
  ITI_PRINT1A("    ICR Register Set at: %#lx\n", iti->address[0]);
  ITI_PRINT1A("    CR Register Set  at: %#lx\n", iti->address[1]);
  
  switch (DEVICE_ID)
    {
    case ITI_ICPCI:
      ITI_PRINT1A("    AM Register Set  at: %#lx\n", iti->address[2]);
      ITI_PRINT1A("    On Board Memory  at: %#lx.\n", iti->address[3]);
      ITI_PRINT1A("    Revision  : %d.\n", iti->revision & ~ICP_MEMSZ_MASK); 
      break;
    case ITI_PCDIG:
      ITI_PRINT1A("    On Board Memory  at: %#lx\n", iti->address[2]);
      ITI_PRINT1A("    Revision  : %d\n", iti->revision & ~PCD_BRDCFG_MASK); 
      break;
    default:
      return -ENXIO;
    }

  ITI_PRINT1("< iti_pci_init\n");
  return OK;
}

/*
 * iti_map_init
 */

int
iti_map_init (itifg_t *iti, void* data)
{
  int error;

  u_char pcd_brdcfg;

  size_t vcr_reg_size;

  u_long align_div[NUM_REGS] = { 0UL, 0UL };
  long align_mod[NUM_REGS] = { 0L, 0L };

  ITI_PRINT1("> iti_map_init\n");
  
  switch (DEVICE_ID)
    {
    case ITI_ICPCI:
      /* memory size is decoded in revision id */
      iti->big_mem = (iti->revision & ICP_MEMSZ_MASK) >> ICP_MEMSZ_SHIFT;
      iti->revision &= ~ICP_MEMSZ_MASK;
      /* PCVision has always 2Mb of on board memory */
      if (iti->revision > PV_REV_MIN && iti->big_mem)
	{
	  iti_printw ("PCVision with 4Mb memory detected?\n");
	  return -EIO;
	}
      /* acquisition control regs are always mmaped */
      if ((iti->address[2] & ITI_PCI_BASE_ADDR_SPACE) !=
	  ITI_PCI_BASE_ADDR_SPACE_MEMORY)
	{
	  iti_printw ("acquisition module registers not mmaped.\n");
	  return -EIO;
	}
      else
	iti->address[2] &= ITI_PCI_BASE_ADDR_MEM_MASK;
      /* frame-grabber memory too */
      if ((iti->address[3] & ITI_PCI_BASE_ADDR_SPACE) !=
	  ITI_PCI_BASE_ADDR_SPACE_MEMORY)
	{
	  iti_printw ("fgrabber on board memory not available.\n");
	  return -EIO;
	}
      else
	iti->address[3] &= ITI_PCI_BASE_ADDR_MEM_MASK;
      /* module register set and on board memory are always aligned */
      if ((iti->address[2] % PAGE_SIZE))
	{
	  iti_printw ("acquisition modul regs not aligned.\n");
	  return -EFAULT;
	}
      if ((iti->address[3] % PAGE_SIZE))
	{
	  iti_printw ("on board memory not aligned.\n");
	  return -EFAULT;
	}
      /* control register space is different! */
      vcr_reg_size = ICP_CR_REG_SIZE;
      break;

    case ITI_PCDIG:
      /* memory size is always 'big' */
      iti->big_mem = TRUE;
      /* board config is decoded in revision id */
      pcd_brdcfg = (iti->revision & PCD_BRDCFG_MASK) >>
	PCD_BRDCFG_SHIFT;
#if (0)	/* FIXME: this does not work */
      switch (pcd_brdcfg)
	{
	case PCD_BRDCFG_4MB_RS422:
	  ITI_PRINT1("    PC-DIG configuration is 4MB/RS-422.\n");
	  break;
	case PCD_BRDCFG_4MB_LVDS:
	  ITI_PRINTI("    PC-DIG configuration is 4MB/LVDS.\n");
	  break;
	default:
	  iti_printna ("    PC-DIG board config is not known (%d).\n",
		       pcd_brdcfg);
	}
#endif
      iti->revision &= ~PCD_BRDCFG_MASK;
      /* frame-grabber memory is always mmaped */
      if ((iti->address[2] & ITI_PCI_BASE_ADDR_SPACE) !=
	  ITI_PCI_BASE_ADDR_SPACE_MEMORY)
	{
	  iti_printw ("fgrabber on board memory not available.\n");
	  return -EIO;
	}
      else
	iti->address[2] &= ITI_PCI_BASE_ADDR_MEM_MASK;
      /* on board memory is always aligned */
      if ((iti->address[3] % PAGE_SIZE))
	{
	  iti_printw ("on board memory not aligned.\n");
	  return -EFAULT;
	}
      /* control register space is different! */
      vcr_reg_size = PCD_CR_REG_SIZE;
      break;

    default:
      return -ENXIO;
    }
  
  /* are cr/icr registers mmaped or io-ports */
  if ((iti->address[0] & ITI_PCI_BASE_ADDR_SPACE) ==
      ITI_PCI_BASE_ADDR_SPACE_IO)
    {
      if (IS_PCDIG)
	{
	  iti_printw ("interface control regs are i/o ports.\n");
	  return -EIO;
	}
      iti->address[0] &= ITI_PCI_BASE_ADDR_IO_MASK;
      iti->regs_io = TRUE;
    } 
  else
    {
      iti->address[0] &= ITI_PCI_BASE_ADDR_MEM_MASK;
      iti->regs_io = FALSE;
    }
  
  /* cr and icr have to be accessed in the same way (mmaped or io) */
  if (iti->regs_io)
    {
      if ((iti->address[1] & ITI_PCI_BASE_ADDR_SPACE) !=
	  ITI_PCI_BASE_ADDR_SPACE_IO)
	{
	  iti_printw ("control regs not accessible via i/o ops.\n");
	  return -EIO;
	}
      else
	{
	  iti->address[1] &= ITI_PCI_BASE_ADDR_IO_MASK;
	  ITI_PRINT1("    ICR/CR Registers are i/o ports.\n");
	}
    }
  else
    {
      if ((iti->address[1] & ITI_PCI_BASE_ADDR_SPACE) !=
	  ITI_PCI_BASE_ADDR_SPACE_MEMORY)
	{
	  iti_printw ("control registers not mmaped.\n");
	  return -EIO;
	}
      else
	{
	  iti->address[1] &= ITI_PCI_BASE_ADDR_MEM_MASK;
	  ITI_PRINT1("    ICR/CR Registers are mmaped\n");
	}
    }
  
  /* align icr/cr register sets to page boundaries (needed for some bios'es) */
  if (!iti->regs_io)
    {
      if ((align_mod[ICR_REGS] = iti->address[0] % PAGE_SIZE))
	iti_printi ("interface control regs not aligned.\n");
      align_div[ICR_REGS] = ROUND_DOWNTO_PAGE(iti->address[0]);
      if (align_div[ICR_REGS] !=
	  ROUND_DOWNTO_PAGE(iti->address[0] + ICR_REG_SIZE - sizeof(u_long)))
	iti_printi ("icr reg-space crosses page boundary.\n");
      if ((align_mod[CR_REGS] = iti->address[1] % PAGE_SIZE))
	iti_printi ("control regs not aligned.\n");
      align_div[CR_REGS] = ROUND_DOWNTO_PAGE(iti->address[1]);
      if (IS_ICPCI)
	if (align_div[CR_REGS] !=
	    ROUND_DOWNTO_PAGE(iti->address[1] + vcr_reg_size-sizeof(u_short)))
	  iti_printi ("cr reg-space crosses page boundary.\n");
    }
  
  if ((error = iti_os_map_init (iti, data, align_div, align_mod)))
    return error;
  
  /* debugging output of remapped kernel space areas */
  if (!iti->regs_io)
    {
      ITI_PRINT3A("    Remapping (ICR) at: 0x%p\n", iti->v_icr_base);
      ITI_PRINT3A("                offset: %#lx\n", align_mod[ICR_REGS]);
      ITI_PRINT3A("                [%ld page(s)]\n",
		  ROUND_UPTO_PAGE(align_mod[ICR_REGS] +
				  ICR_REG_SIZE) / PAGE_SIZE);
      ITI_PRINT3A("    Remapping (CR)  at: 0x%p\n", iti->v_cr_base);
      ITI_PRINT3A("                offset: %#lx\n", align_mod[CR_REGS]);
      ITI_PRINT3A("                [%ld page(s)]\n",
		  ROUND_UPTO_PAGE(align_mod[CR_REGS] +
				  vcr_reg_size) / PAGE_SIZE);
    }

  if (IS_ICPCI)
    {
      ITI_PRINT3A("    Remapping (AM)  at: 0x%p\n", iti->v_mod_base);
      ITI_PRINT3A("                [%ld page(s)]\n",
		  ICP_MOD_MAP_SIZE / PAGE_SIZE);
    }
  
  ITI_PRINT3A("    Remapping (BM)  at: 0x%p\n", iti->v_mem_base);
  switch (DEVICE_ID)
    {
    case ITI_ICPCI:
      ITI_PRINT3A("                [%ld page(s)]\n",
		  iti->big_mem ?
		  (ICP_MEM_MAP_SIZE << 1) / PAGE_SIZE :
		  ICP_MEM_MAP_SIZE / PAGE_SIZE);
      break;
    case ITI_PCDIG:
      ITI_PRINT3A("                [%ld page(s)]\n",
		  PCD_MEM_MAP_SIZE / PAGE_SIZE);
      break;
    default:
      return -ENXIO;
    }
      
  /* fixup by access method / unaligned access */
  if (iti->regs_io)
    {
      iti->v_icr_base = NULL;
      iti->v_cr_base = NULL;
      iti->io_icr_base = iti->address[0];
      iti->io_cr_base = iti->address[1];
    }
  else
    {
      iti->v_icr_base += align_mod[ICR_REGS] / sizeof(u_long);
      iti->v_cr_base += align_mod[CR_REGS];
      iti->io_icr_base = 0UL;
      iti->io_cr_base = 0UL;
    }

  /* assign the different register spaces */
  if (IS_PCDIG)
    {
      iti->pcd_bi_reg = (struct pcd_bi_reg_t*)(iti->v_cr_base);
                                              /* + 0 * PCD_CR_BANK_SIZE; */
      iti->pcd_ca_reg = (struct pcd_ca_reg_t*)(iti->v_cr_base +
					       1 * PCD_CR_BANK_SIZE);
      iti->pcd_oct_table = (pcd_octdata_t*)(iti->v_cr_base +
					    2 * PCD_CR_BANK_SIZE);
      iti->pcd_dma_table = (pcd_dmadata_t*)(iti->v_cr_base +
					    3 * PCD_CR_BANK_SIZE);
      iti->pcd_lut0_table = (u_long*)(iti->v_cr_base +
				      4 * PCD_CR_BANK_SIZE);
      iti->pcd_lut1_table = (u_long*)(iti->v_cr_base +
				      5 * PCD_CR_BANK_SIZE);
      /* pcd_uart_reg = (struct pcd_uart_ret_t*)(iti->v_cr_base + */
                                                /*  6 * PCD_CR_BANK_SIZE); */
      iti->pcd_fp_reg = (struct pcd_fp_reg_t*)(iti->v_cr_base +
					       7 * PCD_CR_BANK_SIZE);

      ITI_PRINT3A("    BI Register Set at: 0x%p\n", iti->pcd_bi_reg);
      ITI_PRINT3A("    CA Register Set at: 0x%p\n", iti->pcd_ca_reg);
      ITI_PRINT3A("    OC Table        at: 0x%p\n", iti->pcd_oct_table);
      ITI_PRINT3A("    DMA Table       at: 0x%p\n", iti->pcd_dma_table);
      ITI_PRINT3A("    LU Table 0      at: 0x%p\n", iti->pcd_lut0_table);
      ITI_PRINT3A("    LU Table 1      at: 0x%p\n", iti->pcd_lut1_table);
      ITI_PRINT3A("    FPGA Download   at: 0x%p\n", iti->pcd_fp_reg);
    }

  ITI_PRINT1("< iti_map_init\n");    
  return OK;
}

/*
 * a board can only be under control of ONE module 
 */

int
iti_alloc_board (itifg_t *iti)
{
  int iti_used_version;
  extern int iti_version_code;

  ITI_PRINT1("> iti_alloc_board\n");    

  if ((iti_used_version = icr_get_mbox1 (iti)) != ICR_MBOX_INIT)
    {
      iti_printn ("This board is already under control\n");
      iti_printna ("of another device driver (v%d.%d.%d%c).\n",
		   (iti_used_version & 0xFF000000) >> 24,
		   (iti_used_version & 0x00FF0000) >> 16,
		   (iti_used_version & 0x0000FF00) >> 8,
		   (iti_used_version & 0x000000FF) + '@');
      return ERROR;
    }
  
  icr_set_mbox1 (iti, iti_version_code);
  iti->used_by_me = TRUE;

  ITI_PRINT1("< iti_alloc_board\n");    
  return OK;
}

void
iti_free_board (itifg_t *iti)
{
  if (iti->used_by_me)
    {
      icr_set_mbox1 (iti, ICR_MBOX_INIT);
      iti->used_by_me = FALSE;
    }
}

/*
 * iti_cap_init
 */

int
iti_cap_init (itifg_t *iti)
{

  int error;

  ITI_PRINT1("> iti_cap_init\n");

  /* test for various icpci hardware bugs */
  if (IS_ICPCI)
    {
      if (iti->revision == 0)
	/* AMCC Rev0 Bug */
	{
	  iti_printw ("AMCC PCI Matchmaker Rev. 0 detected!\n");
	  iti_printw ("This is no longer supported!\n");
	  return -ENXIO;
	}
      else
	{
	  /* AMCC Rev1 Bug (S5933Q vs. S5933QB) */
	  if ((iti->status & ICP_LEVEL_MASK) >> ICP_LEVEL_SHIFT)
	    {
	      iti_printw ("AMCC PCI Matchmaker Rev. 1Q detected!\n");
	      iti_printw ("This is no longer supported!\n");
	      return -ENXIO;
	    }
	}
    }

  /* test for successfully loaded FPGA's */
  if (IS_ICPCI)
    iti->fpga_ok = TRUE;  /* here we don't have to load a FPGA */
  else
    if (IS_PCDIG)
      {
	if (pcd_get_fpgadone (iti))
	  iti->fpga_ok = TRUE;
	else
	  iti->fpga_ok = FALSE;
      }
  
  /* test, if board is already used by a driver */
  if (iti_alloc_board (iti))
    return -EBUSY;

  /* initialize the base register sets */
  if ((error = iti_icr_init (iti)))
    {
      iti_printw ("Init. of Interface Control Regs. failed.\n");
      return error;
    }
  if ((error = iti_cr_init (iti)))
    {
      iti_printw ("Init. of Control Regs. failed.\n");  
      return error;
    }

  /* autodetect acquisition module and initialize it */
  switch (DEVICE_ID)
    {
    case ITI_ICPCI:
      if (!cr_get_amp(iti))
	{
	  iti->module = icp_get_modid (iti);
	  switch (iti->module)
	    {
	    case ICP_AMFA:
	      iti_printn ("AM-FA Acquisition Module found.\n");
	      iti_printn ("Not supported yet!.\n");
	      return -ENXIO;
	      break;
	    case ICP_AMC1:
	      iti_printn ("AM-C1 Acquisition Module found.\n");
	      iti_printn ("Not supported yet!.\n");
	      return -ENXIO;
	      break;
	    case ICP_AMMTD:
	      iti_printn ("AM-MTD Acquisition Module found.\n");
	      iti_printn ("Not supported yet!.\n");
	      return -ENXIO;
	      break;
	    case ICP_AMCMP:
	      iti_printn ("AM-STD/COMP Acquisition Module found.\n");
	      iti_printn ("Not supported yet!.\n");
	      return -ENXIO;
	      break;
	    case ICP_AMRGB:
	      iti_printn ("AM-STD/RGB Acquisition Module found.\n");
	      iti_printn ("Not supported yet!.\n");
	      return -ENXIO;
	      break;
	    case ICP_AMVS:
	      iti_printi ("AM-VS Acquisition Module found.\n");
	      if ((error = icp_amvs_startup (iti)))
		{
		  iti_printw ("Startup of AM-VS Regs. failed.\n");
		  return error;
		}
	      iti->cameras = VS_CAMERAS_MAX;
	      break;
	    case ICP_AMDG:
	      iti_printi ("AM-DIG Acquisition Module found.\n");
	      if ((error = icp_amdg_startup (iti)))
		{
		  iti_printw ("Startup of AM-DIG Regs. failed.\n");
		  return error;
		}
	      iti->cameras = DG_CAMERAS_MAX;
	      break;
	    case ICP_AMPV:
	      iti_printi ("PCVision Acquisition Module found.\n");
	      iti->pv_reg = (struct pv_reg_t*)(iti->v_mod_base);
	      if ((error = icp_ampv_startup (iti)))
		{
		  iti_printw ("Startup of PCVision Regs. failed.\n");
		  return error;
		}
	      iti->cameras = PV_CAMERAS_MAX;
	      break;
	    default:
	      iti_printna ("Unknown Module Type (%d) found.\n",
			   iti->module);
	      return -ENXIO;
	    }
	}
      else
	{
	  iti_printw ("No Acquisition Module installed.\n");
	  return -ENXIO;
	}
      break;

    case ITI_PCDIG:
      iti->module = ITI_VMPCD;
      if ((error = pcd_viam_startup (iti)))
	{
	  iti_printw ("Startup of PC-DIG Regs. failed.\n");
	  return error;
	}
      iti->cameras = PCD_CAMERAS_MAX;
      break;

    default:
      return -ENXIO;
    }

  ITI_PRINT1("< iti_cap_init\n");
  return OK;
}

/*
 * iti_dev_init
 */

int
iti_dev_init (itifg_t *iti)
{

  int error;
  
  ITI_PRINT1("> iti_dev_init\n");
  
  switch (DEVICE_ID)
    {
    case ITI_ICPCI:
      switch (MODULE_ID)
	{
	case ICP_AMVS:
	  iti_memcpy (&iti->dev.vs_cam, &vs_def_cam, sizeof(struct vs_cam_t));
	  if ((error = icp_amvs_init (iti)))
	    {
	      iti_printw ("Init. of AM-VS Regs. failed.\n");
	      return error;
	    }
	  break;
	case ICP_AMDG:
	  iti_memcpy (&iti->dev.dg_cam, &dg_def_cam, sizeof(struct dg_cam_t));
	  if ((error = icp_amdg_init (iti)))
	    {
	      iti_printw ("Init. of AM-DIG Regs. failed.\n");
	      return error;
	    }
	  break;
	case ICP_AMPV:
	  iti_memcpy (&iti->dev.pv_cam, &pv_def_cam, sizeof(struct pv_cam_t));
	  if ((error = icp_ampv_init (iti)))
	    {
	      iti_printw ("Init. of PCVision Regs. failed.\n");
	      return (error);
	    }
	  break;
	default:
	  return -ENXIO;
	}
      break;
      
    case ITI_PCDIG:
      iti_memcpy (&iti->dev.pcd_cam, &pcd_def_cam, sizeof(struct pcd_cam_t));
      if ((error = pcd_viam_init (iti)))
	{
	  iti_printw ("Init. of PC-DIG Regs. failed.\n");
	  return error;
	}
      break;
	
    default:
      return -ENXIO;
    }
  
  ITI_PRINT1("< iti_dev_init\n");
  return OK;
}

/*
 * iti_detach
 */

int
iti_detach (void)
{
  int board, snap;
  itifg_t *iti;

  ITI_PRINT0("> iti_detach\n");
  
  board = 0;
  while (itifg[board] != NULL)
    {
      iti = itifg[board];
      for (snap = 0; snap < ITI_SNAPS_MAX; snap++)
	if (iti->sglist[snap] != NULL)
	  {
	    ITI_PRINT3A("     sglist[%d]: 0x%p\n", snap, iti->sglist[snap]);
	    iti_free (iti->sglist[snap], ITI_UNPAGED);
	  }
      /* map is handled in the os specific function */
      ITI_PRINT3A("        iti: 0x%p\n", iti);
      iti_free (iti, ITI_UNPAGED);
      board++;
    }
  
  ITI_PRINT0("< iti_detach: OK\n");  
  return OK;
}

/*
 * iti_attach
 */

int
iti_attach (int board, void *data)
{
  int error;
  itifg_t *iti;
  
  ITI_PRINT0A("> iti_attach %d\n", board);

  /* allocate the global structure */
  ITI_PRINT3A("    iti: %d\n", sizeof (itifg_t));
  itifg[board] = (itifg_t*)iti_malloc (sizeof (itifg_t), ITI_UNPAGED);
  if (itifg[board] == NULL)
    return -ENOMEM;
  ITI_PRINT3A("    iti: 0x%p\n", itifg[board]);
  iti_memset (itifg[board], 0, sizeof(itifg_t));
  
  iti = itifg[board];

  /* initialize data structures */
  if ((error = iti_os_init (iti)) < 0)
    goto attach_error;
  if ((error = iti_pci_init (iti, data)) < 0)
    goto attach_error;
  if ((error = iti_map_init (iti, data)))
    goto attach_error;
  if ((error = iti_cap_init (iti)))
    goto attach_error;
  if ((error = iti_dev_init (iti)))
    goto attach_error;
  if ((error = iti_fmt_init (iti)))
    goto attach_error;
  if ((error = iti_buf_init (iti)))
    goto attach_error;

  /* reset statistics */
  iti->acc.frames_captured = 0UL;
  iti->acc.frames_transfered = 0UL;
  iti->acc.frames_timedout = 0UL;
  iti->acc.frames_get_by_blocked = 0UL;
  iti->acc.frames_get_by_signal = 0UL;
  iti->acc.frames_get_by_select = 0UL;
  iti->acc.events_captured = 0UL;
  iti->acc.bytes_valid = 0UL;
  iti->acc.bytes_lost = 0UL;
  
  goto attach_done;
  
 attach_done:
  ITI_PRINT0A("< iti_attach %d: OK\n", board);
  return OK;
  
 attach_error:
  ITI_PRINT0A("< iti_attach %d: error %d\n", board, error);
  return error;
}

